/**
 *  VUCAKO Bench in Java: main program.
 *   @author  PE
 *   @version 0.007 (2001/06/26)
 *   @since   0.006
 */
public class Bench {

    /** Bit-mask of all implemented tests. */
    protected final static int MAX_TEST_MASK = 0x7FFF;

    /** Default tests. */
    protected final static int DEFAULT_TEST_MASK = 0x4FFB;

    /**
     *  Formats integer in &quot;%.3f&quot; C-format.
     *   @param  value Number to be formatted.
     *   @return String representation of the given number.
     */
    protected static String formatSec ( long value ) {
        return( Long.toString(value/1000L) + "." +
                Long.toString((value/100L)%10L) +
                Long.toString((value/10L)%10L) +
                Long.toString(value%10L) );
    }

    /**
     *  Runs the given benchmark.
     *   @param  b Initialized <code>Benchamrk</code> instance.
     */
    protected static void runTest ( Benchmark b ) {
        long elapsed = b.runTest();
        System.out.println( b.testResult() + ", elapsed time: " + formatSec(elapsed) + " sec" );
    }

    /**
     *  Main jBench routine.
     *   @param  args Arguments passed to the program.
     */
    public static void main ( String[] args ) {

        System.out.println("VUCAKO jBench ver 1.007 (c) 2001 Josef.Pelikan@mff.cuni.cz (OpenSource)");

        int testMask = DEFAULT_TEST_MASK;
        int argIndex;
        boolean alt = false;
        for ( argIndex = 0; argIndex < args.length; argIndex++ ) {
            if ( args[argIndex].equalsIgnoreCase("alt") )
                alt = true;
            else
                testMask = Integer.parseInt(args[argIndex],16);
            }
        testMask &= MAX_TEST_MASK;
        System.out.println( "Running tests: " + Integer.toString(testMask,16) +
                            (alt ? " alt" : "") );

            // 1 - Eratosthen
        if ( (testMask & 0x0001) != 0 )
            runTest( new BenchmarkEratosthen() );

            // 2 - Image transposition (16MB image)
        if ( (testMask & 0x0002) != 0 )
            runTest( new BenchmarkTranspose(2,alt) );

            // 3 - Image transposition (64MB image)
        if ( (testMask & 0x0004) != 0 )
            runTest( new BenchmarkTranspose(3,alt) );

            // 4 - Arbitrage sequence lookup
        if ( (testMask & 0x0008) != 0 )
            runTest( new BenchmarkArbitrage() );

            // 5 - Needle-throwing simulation
        if ( (testMask & 0x0010) != 0 )
            runTest( new BenchmarkNeedle() );

            // 6 - Fast memory-copy
        if ( (testMask & 0x0020) != 0 )
            runTest( new BenchmarkMemoryCopy() );

            // 7 - Monte-Carlo form-factor
        if ( (testMask & 0x0040) != 0 )
            runTest( new BenchmarkFormFactor() );

            // 8 - Merge-sort on disk
        if ( (testMask & 0x0080) != 0 )
            runTest( alt ? new BenchmarkMergeSortAlt() : new BenchmarkMergeSort() );

            // 9 - Wavelet transform in memory
        if ( (testMask & 0x0100) != 0 )
            runTest( new BenchmarkWaveletMemory() );

            // 10 - Wavelet transform on disk
        if ( (testMask & 0x0200) != 0 )
            runTest( new BenchmarkWaveletDisk() );

            // 11 - Adaptive K-D tree on disk
        if ( (testMask & 0x0400) != 0 )
            runTest( new BenchmarkAKDTree() );

            // 12 - Quick-sort in virtual memory (64MB array)
        if ( (testMask & 0x0800) != 0 )
            runTest( new BenchmarkQuickSort(12) );

            // 13 - Quick-sort in virtual memory (256MB array)
        if ( (testMask & 0x1000) != 0 )
            runTest( new BenchmarkQuickSort(13) );

            // 14 - Quick-sort in virtual memory (1GB array)
        if ( (testMask & 0x2000) != 0 )
            runTest( new BenchmarkQuickSort(14) );

            // 15 - Garbage collection in 64MB RAM
        if ( (testMask & 0x4000) != 0 )
            runTest( new BenchmarkGC() );

    }

}
